"""Module to hold every utility of the library."""

from typing import List, Optional

import serial
from serial.tools.list_ports_common import ListPortInfo

from iic2343.constants import (
    BAUD_RATE,
    BOTTOM_ADDRESS,
    BYTE_SIZE,
    PARITY,
    RTS_CTS,
    STOP_BITS,
    TOP_ADDRESS,
    WORD_LENGTH,
    XON_XOFF,
)
from iic2343.errors import CannotOpenPortError, InvalidPortError, NoPortsPresentError


def configure_port(port: serial.Serial) -> serial.Serial:
    """Configures the required parameters of a port."""
    port.baudrate = BAUD_RATE
    port.bytesize = BYTE_SIZE
    port.parity = PARITY
    port.rtscts = RTS_CTS
    port.stopbits = STOP_BITS
    port.xonxoff = XON_XOFF
    return port


def open_port(port: serial.Serial) -> None:
    """Handles the opening of a port."""
    port.open()
    if port.is_open:
        port.write([0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF])
    else:
        raise CannotOpenPortError("Cannot open requested serial port.")


def close_port(port: serial.Serial) -> None:
    """Handles the closing of a port."""
    port.write([0xFF])
    port.close()


def validate_port_selection(
    port_number: Optional[int], os_ports: List[ListPortInfo]
) -> None:
    """
    Checks if a port number is valid for an array of available ports.
    If the port number is invalid, raises an error.
    """
    if not os_ports:
        raise NoPortsPresentError("No serial ports present.")
    if len(os_ports) > 1 and port_number is None:
        raise InvalidPortError(
            f"{len(os_ports)} serial ports present, please select one."
        )
    if port_number is not None and not -1 < port_number < len(os_ports):
        raise InvalidPortError(
            f"Invalid port number '{port_number}', {len(os_ports)} serial "
            "ports present, please select a valid one."
        )


def can_be_written(port: serial.Serial, address: int, word: bytearray) -> bool:
    """Checks if a word can be written on an address through a given port."""
    if address < BOTTOM_ADDRESS:
        return False
    if address >= TOP_ADDRESS:
        return False
    if len(word) != WORD_LENGTH:
        return False
    if not port.is_open:
        return False
    return True


def write_to_port(port: serial.Serial, address: int, word: bytearray) -> int:
    """
    Writes some data into an address through a port, it assumes that the
    data can be written.
    """
    add = (address << 4).to_bytes(2, "big")
    payload = bytearray(
        [
            0xAA,
            0xAA,
            add[0],
            add[1] | (word[0] & 0xF),
            word[1],
            word[2],
            word[3],
            word[4],
            0xAA,
        ]
    )
    return port.write(payload)



"""Core module for the package. It holds the main object to be used."""

from typing import Optional

import serial
from serial.tools import list_ports

from iic2343.utils import (
    can_be_written,
    close_port,
    configure_port,
    open_port,
    validate_port_selection,
    write_to_port,
)


class Basys3:

    """
    Encapsulates the behaviour required for the data to be written to
    the Basys3 board.
    """

    def __init__(self) -> None:
        self.__port = serial.Serial()

    def begin(self, port_number: Optional[int] = None) -> None:
        """Configure and initialize the port to be used."""
        os_ports = list_ports.comports()
        validate_port_selection(port_number, os_ports)
        self.__port = configure_port(self.__port)
        self.__port.port = os_ports[port_number or 0].device
        open_port(self.__port)

    def end(self) -> None:
        """Close the port."""
        close_port(self.__port)

    def write(self, address: int, word: bytearray) -> int:
        """Write to the initialized port."""
        if not can_be_written(self.__port, address, word):
            return 0
        return write_to_port(self.__port, address, word)



"""Module used to hold potential errors present on the client code."""


class BaseIIC2343Error(Exception):
    """
    A base error class to be able to handle every error thrown by the
    library at the same time.
    """


class CannotOpenPortError(BaseIIC2343Error):
    """Error for when a port cannot be opened."""


class InvalidPortError(BaseIIC2343Error):
    """Error for when no ports are selected or an invalid port is selected."""


class NoPortsPresentError(BaseIIC2343Error):
    """Error for when no ports can be found on the machine."""



"""This module holds every constant for the library."""


# Port configurations
BAUD_RATE = 115200
PARITY = "N"
BYTE_SIZE = 8
STOP_BITS = 2
XON_XOFF = 0
RTS_CTS = 0


# Write constants
BOTTOM_ADDRESS = 0
TOP_ADDRESS = 2 ** 12
WORD_LENGTH = 5



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iic2343']

package_data = \
{'': ['*']}

install_requires = \
['pyserial>=3.5,<4.0']

setup_kwargs = {
    'name': 'iic2343',
    'version': '0.1.1',
    'description': 'Write to the Basys3 directly.',
    'long_description': "# IIC2343\n\n## Installation\n\nInstall using `pip`!\n\n```sh\npip install iic2343\n```\n\n## Usage\n\nTo use the library, import the `Basys3` object directly and use the `begin`, `write` and `end` methods!\n\n```python\nfrom iic2343 import Basys3\n\ninstance = Basys3()\n\ninstance.begin(port_number=2)  # port_number is optional\ninstance.write(1, bytearray([0x00, 0x00, 0x10, 0x16, 0x01]))\ninstance.write(2, bytearray([0x00, 0x00, 0x00, 0x18, 0x03]))\ninstance.write(3, bytearray([0x00, 0x00, 0x20, 0x18, 0x03]))\ninstance.write(4, bytearray([0x00, 0x00, 0x00, 0x20, 0x00]))\ninstance.end()\n```\n\nHere, a `Basys3` instance has 3 methods:\n\n### `begin`\n\nThe method receives an optional `port_number` parameter (in needs to be an `int`). If the parameter is not present and there is only one available serial port on your machine, the `Basys3` instance will use that serial port. Otherwise, it will raise an exception. The method initializes a port to `write` to.\n\n### `write`\n\nThe method receives an `address` parameter (an `int`) and a `word` parameter (a `bytearray`). It then attempts to write the `word` on the specified `address`. If the `Basys3` instance fails, it returns a `0`. Otherwise, it returns an `int`.\n\n### `end`\n\nThe method receives no parameters, and simply closes the port initialized on the `begin` method.\n\n\n## Developing\n\nThis library uses `PyTest` as the test suite runner, and `PyLint`, `Flake8`, `Black`, `ISort` and `MyPy` as linters. It also uses `Poetry` as the default package manager.\n\nThe library includes a `Makefile` that has every command you need to start developing. If you don't have it, install `Poetry` using:\n\n```sh\nmake get-poetry\n```\n\nThen, create a virtualenv to use throughout the development process, using:\n\n```sh\nmake build-env\n```\n\nActivate the virtualenv using:\n\n```sh\n. .venv/bin/activate\n```\n\nDeactivate it using:\n\n```sh\ndeactivate\n```\n\nTo add a new package, use `Poetry`:\n\n```sh\npoetry add <new-package>\n```\n\nTo run the linters, you can use:\n\n```sh\n# The following commands auto-fix the code\nmake black!\nmake isort!\n\n# The following commands just review the code\nmake black\nmake isort\nmake flake8\nmake mypy\nmake pylint\n```\n\nTo run the tests, you can use:\n\n```sh\nmake tests\n```\n\n## Releasing\n\nTo make a new release of the library, `git switch` to the `master` branch and execute:\n\n```sh\nmake bump! minor\n```\n\nThe word `minor` can be replaced with `patch` or `major`, depending on the type of release. The `bump!` command will bump the versions of the library, create a new branch, add and commit the changes. Then, just _merge_ that branch to `master`. Finally, execute a _merge_ to the `stable` branch. Make sure to update the version before merging into `stable`, as `PyPi` will reject packages with duplicated versions.\n",
    'author': 'Daniel Leal',
    'author_email': 'dlleal@uc.cl',
    'maintainer': 'Daniel Leal',
    'maintainer_email': 'dlleal@uc.cl',
    'url': 'https://github.com/daleal/iic2343',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
